#include "WFrame.h"

// EdgeManager.cpp
// Copyright 1998, William Bull

//............................................................
// In general this should be used in conjunction
// with something which calls Layout(). If you're dealing
// with a static layout this may be useful as a
// stand-alone class. Presently I incorporate it into
// WLayoutWindow for experimenting with real time
// layout management.
//
// Other layout managers can sit on top of this class
// for flow layout, tk packing, java'ish awt layouts,
// etc. These will likely be classes which have an
// instance of WEdgeManger which they communicate to
// for an initial layout then get out of the way for
// it to do whatever arrangment they just described.

//............................................................
// See: WFrame.h
//
// This uses WFrames to describe the positioning
// for any BView. Multiple BViews may use the same
// WFrame. This will allow for tab view classes
// and other hiding/showing scenerios to be easier. 

#pragma once

#include "WFrame.h"

typedef struct 
{
	WFrame			*frame;
	BView			*view;
	BRect			newFrame;
	bool			isLeftSet,
					isTopSet,
					isRightSet,
					isBottomSet;
} WFrameEntry;

class WEdgeManager
{
public:
						WEdgeManager();
	virtual				~WEdgeManager();
//-------------------------------------------------- public
	virtual void		Manage(BView *view,
							   WFrame *frame);
	virtual void		StopManaging(const WFrame *frame);
	virtual void		StopManaging(const BView *view);
	virtual void		Remove(const BView *view);
	virtual	void		Layout(const BRect& frame);

//... the thing that does all the work (called by layout) ...

	virtual void		Arrange(WFrameEntry *frame,
								const ulong edge,
								const BRect& parentFrame);

//------------------------------------------------- internal
	WFrameEntry*		FindEntry(const BView *view) const;
	WFrameEntry*		FindEntry(const WFrame *frame) const;
	void				Move(BView *view,
								const BRect& moveRect);
	BRect				ParentBounds(const BView *view) const;
	BList				*entries;
	
/*....................unimplemented.........................

	This was supposed to allow for localize rearrangement
	where a single view could be told to adjust itself.
	Unfortunately many dependencies makes its usefulness
	questionable:
	
	virtual	void		Layout(BView *view,
								const BRect& frame);
*/
};