// Frame.cpp
// Copyright 1998, William Bull

//-------------------------------------------------------------
// Ok, here's the basic premise:
//
// A container is generally divided into regions where components
// are added. These regions may shift around and be reused by
// lots of sets of components. As you resize a window some
// regions may proportionally adjust, others remain
// in a specific spot. I call these regions WFrames.
//
// This is important to remember:
// 
// A WFrame doesn't actually exist on the screen. It has no
// literal coordinate values. It can only resolve an edge value
// when combined with a BView. Building edge dependencies
// for BViews must be described in terms of other BViews, not
// WFrames. In turn those BViews may use a WFrame to associate
// its edges to itself, siblings or its parent.
//-------------------------------------------------------------

//-------------------------------------------------------------
// Examples for Set:
//
// Set(EDGE_LEFT, USE_NONE)
// Set(EDGE_LEFT, USE_DELTA, -10)

// Set(EDGE_LEFT, USE_PARENT, USE_LITERAL, 10)
// Set(EDGE_LEFT, USE_PARENT, USE_PERCENTAGE, 50)

// Set(EDGE_LEFT, USE_SIBLING, BView*, EDGE_ADJACENT, 5)
// Set(EDGE_LEFT, USE_SIBLING, BView*, EDGE_OPPOSITE, 5)
// Set(EDGE_LEFT, USE_SIBLING, BView*, USE_PERCENTAGE, 50)
//-------------------------------------------------------------

#pragma once

enum {
	// which style
	USE_LITERAL = 0,
	USE_PERCENT,
	
	USE_DELTA,	// assume dependency is USE_SELF
	USE_NONE,	// assume dependency is USE_SELF
	
	// which dependency
	USE_PARENT,
	USE_SIBLING,
	
	// which edge
	EDGE_LEFT,
	EDGE_TOP,
	EDGE_RIGHT,
	EDGE_BOTTOM,
	EDGE_OPPOSITE,
	EDGE_ADJACENT
};

class WFrame
{
public:
			WFrame(BView *which = 0);
virtual		~WFrame();
//------------------------------------------------ public
	void	Set(ulong whichEdge,
				ulong whichRelation, /* assume USE_SELF */
				float howMuch);
	void	Set(ulong whichEdge,
				ulong whichRelation,      /* USE_PARENT */
				ulong whichViewRelation,
				float howMuch);
	void	Set(ulong whichEdge,
				ulong whichRelation,     /* USE_SIBLING */
				BView *whichView,
				ulong whichViewRelation,
				float howMuch);
//------------------------- data necessary for managers

// A parameter for the various positioning methods
float		left_value, right_value,
			top_value, bottom_value,
// A parameter for a min/max size (<0 doesn't care)
			min_height,
			min_width,
			max_height,
			max_width;
// A description of how to position an edge
ulong		left_relation,
			right_relation,
			top_relation,
			bottom_relation,
// A dependent relational edges (adj, opp)
			left_style,
			top_style,
			right_style,
			bottom_style;
// A pointer to a relational view
BView		*left_view,
			*right_view,
			*top_view,
			*bottom_view;
};