// EdgeManager.cpp
// Copyright 1998, William Bull

/*
	Note: This code is under redevelopment and has not been cleaned up.
	Please excuse any vulgarities or painfully obvious lack of coding style
	on my behalf. 
*/


/*
		WARNING
		
		Potential bug - Bounds() calls are not OffsetTo(0,0). Thus
		components arranged within ciews which have ScrollTo calls
		adjusting the viewable regions may adjust wrong. This has
		not been tested - this code works under "normal" cases so
		I have not made the OffsetTo change. (Fear of mucking up
		something that appears to be working. :)


		Ok, it appears that this is a problem. This version of
		the edge manager now has changes to fix it.
*/

#include "WEdgeManager.h"

int32 ROUND_ERROR_CORRECT = .1;

//............................................................
// Constructor
// Not much going on here... i try to use
// as little dynamic memory as possible.

WEdgeManager::WEdgeManager()
{
	entries = new BList();
}

//............................................................
// Destructor
// Just cleaning up the BList. Everything within the
// entries are simple references to other objects.
// (Permits sharing, managability, etc.)

WEdgeManager::~WEdgeManager()
{
	WFrameEntry *temp;
	
	for (int i = entries->CountItems(); i > 0; i--)
	{
		temp = (WFrameEntry*)(entries->ItemAt(i-1));
		delete temp;
	}
}

//............................................................
// Manage()
// Construct an entry if necessary to desribe which WFrame
// this view is to be managed by. The WFrame will dictate the
// resizing behaviour and position it has due to relationships.

void WEdgeManager::Manage(BView *view, WFrame *frame)
{
	if (frame == NULL) return;
	if (view == NULL) return;
	
	WFrameEntry *newFrame = FindEntry(view);
	if (newFrame == NULL) {
		newFrame = new WFrameEntry();
	}
	
	newFrame = new WFrameEntry;
	newFrame->frame = frame;
	newFrame->view = view;
	newFrame->newFrame = view->Frame();
	newFrame->isLeftSet = false;
	newFrame->isRightSet = false;
	newFrame->isTopSet = false;
	newFrame->isBottomSet = false;
	
	entries->AddItem((void*)(newFrame));
}

//............................................................
// FindEntry()
// Moves through the BList looking for an entry with a
// corresponding view. Returns the entry or NULL if it fails.

WFrameEntry* WEdgeManager::FindEntry(const BView *view) const
{
	WFrameEntry *frame;
	
	for (int i = 0; i < entries->CountItems(); i++)
	{
		frame = (WFrameEntry*)entries->ItemAt(i);
		if (frame->view == view) {
			return frame;
		}
	}	
	return NULL;
}

//............................................................
// FindEntry()
// Moves through the BList looking for the first entry which
// uses the WFrame. Returns the entry or NULL if it fails.

WFrameEntry* WEdgeManager::FindEntry(const WFrame *frame) const
{
	WFrameEntry *f;
	for (int i = 0; i < entries->CountItems(); i++)
	{
		f = (WFrameEntry*)entries->ItemAt(i);
		if (f->frame == frame) return f;
	}	
	return NULL;
}

//............................................................
// StopManaging()
// If for some reason you no longer want the manger to have
// any reference to a WFrame, this removes all entries with
// BViews associated to the specified WFrame.

void WEdgeManager::StopManaging(const WFrame *frame)
{
	WFrameEntry *entry = FindEntry(frame);
	
	while (entry != NULL)
	{
		entries->RemoveItem((void*)entry);
		delete entry;
		entry = FindEntry(frame);
	}
}

//............................................................
// StopManaging()
// If for some reason you no longer want to adjust a BView
// this will find the BView entries which are associated to
// any WFrames. Normally there should only ever be one entry
// per BView but this checks for multiple entries and removes
// those too.
//
// Note: Other entries of WFrames referring to BViews may
//       still exist. Only direct dependencies are
//       removed, ie. ones this manager has been told to manage.

void WEdgeManager::StopManaging(const BView *view)
{
	WFrameEntry *entry = FindEntry(view);
	
	while (entry != NULL)
	{
		entries->RemoveItem((void*)entry);
		delete entry;
		entry = FindEntry(view);
	}

}

//............................................................
// Remove()
// All associations to the given BView are removed. Frames
// which had dependent edges now have edges which are USE_NONE.
//
void WEdgeManager::Remove(const BView *view)
{
	WFrameEntry *entry = FindEntry(view);
	
	while (entry != NULL)
	{
		entries->RemoveItem((void*)entry);
		delete entry;
		entry = FindEntry(view);
	}

	for (int i = 0; i < entries->CountItems(); i++)
	{
		entry = (WFrameEntry*)entries->ItemAt(i);
		if (entry->frame->left_view == view) {
			entry->frame->left_view = 0;
			entry->frame->left_relation = USE_NONE;
	 	}
	 	if (entry->frame->top_view == view) {
			entry->frame->top_view = 0;
			entry->frame->top_relation = USE_NONE;
	 	}
	 	if (entry->frame->bottom_view == view) {
			entry->frame->bottom_view = 0;
			entry->frame->bottom_relation = USE_NONE;
	 	}
	 	if (entry->frame->right_view == view) {
			entry->frame->right_view = 0;
			entry->frame->right_relation = USE_NONE;
	 	}
	}	
}

//............................................................
// ParentBounds()
// Returns the bounds of the parent window or view. If neither
// exist, returns the bounds of itself. (Works for my needs...)

BRect WEdgeManager::ParentBounds(const BView *view) const
{
	BRect r;
	if (view->Parent()) {
		r =  view->Parent()->Bounds();
		r.OffsetTo(0,0);
		//printf("Parent is %d,%d x %d,%d.\n", (int)r.left, (int)r.top, (int)r.right, (int)r.bottom);
		return r;
	}
	if (view->Window()) {
		r = view->Window()->Bounds();
		r.OffsetTo(0,0);
		//printf("Parent is %d,%d x %d,%d.\n", (int)r.left, (int)r.top, (int)r.right, (int)r.bottom);
		return r;
	}
	r = view->Bounds();
	r.OffsetTo(0,0);
	//printf("Parent is %d,%d x %d,%d.\n", (int)r.left, (int)r.top, (int)r.right, (int)r.bottom);
	return r;
}

//............................................................
// Move()
// Calls MoveTo() and ResizeBy() to ensure the view fits
// correctly in the given rect.

void WEdgeManager::Move(BView *view, const BRect& rect)
{
	if (view->Parent() == NULL &&
		view->Window() == NULL) return;
	
	float dX = rect.right - rect.left;
	float dY = rect.bottom - rect.top;
	
	//printf("Moving view to %d,%d x %d,%d.\n", (int)rect.left, (int)rect.top,
	//		(int)rect.right, (int)rect.bottom);
	view->MoveTo(rect.left, rect.top);
	view->ResizeTo(dX, dY);
}

//............................................................
// Layout()
// Proceeds through each entry and attempts to adjust each
// edge if it has not already been adjusted. The parentFrame
// is retained for synchronization. Similarly, the moves and
// the reset of the flags attempt to be fairly synchronized.
// This presents a problem when multiple managers begin to
// play with WFrames since there is no locking code anywhere
// and playing with multithreaded layout schemes will likely
// lead to a deadlocked Hell. Perhaps a job to tackle later...
//
// (For now a single manager managing a single WFrame is fine.
//  Possibly fine for multiple managers if the scheduling is
//  done right.)

void WEdgeManager::Layout(const BRect& parentFrame)
{
	static bool flag = true;
	//printf("\n");
	WFrameEntry *newFrame;
	
	for (int i = 0; i < entries->CountItems(); i++)
	{
		newFrame = (WFrameEntry*)entries->ItemAt(i);
		//printf("Layout:\n\tbottom relation=%d\n",newFrame->frame->bottom_relation);
		if (!newFrame->isTopSet) Arrange(newFrame,EDGE_TOP, parentFrame);
		if (!newFrame->isLeftSet) Arrange(newFrame,EDGE_LEFT, parentFrame);
		if (!newFrame->isRightSet) Arrange(newFrame,EDGE_RIGHT, parentFrame);
		if (!newFrame->isBottomSet) Arrange(newFrame,EDGE_BOTTOM, parentFrame);
	}
	
	for (int i = 0; i < entries->CountItems(); i++)
	{
		newFrame = (WFrameEntry*)entries->ItemAt(i);
		//printf("Moving a view.\n");
		Move(newFrame->view, newFrame->newFrame);
	}
	
	for (int i = 0; i < entries->CountItems(); i++)
	{
		newFrame = (WFrameEntry*)entries->ItemAt(i);
		newFrame->isTopSet = false;
		newFrame->isLeftSet = false;
		newFrame->isRightSet = false;
		newFrame->isBottomSet = false;
	}

	if (flag) {
		flag = false;
		Layout(parentFrame);
	} else flag = true;
}

//............................................................
// Layout()
// Will be used to optimize for effeciency in lay out
// situations where only a single view has adjusted its
// state (rather than something like a total resize of the
// window.) It minimizes the layout call to only those
// dependent on the view which has had a change and the
// view itself.
//
// Presently, I'm not sure of the best way to implement
// this. Perhaps a dependency list built into the entry;
// unfortunately that's a lot of overhead. I'll put it
// off until later.
//
//void WEdgeManager::Layout(BView *which, const BRect& parentFrame)
//{
//}

// Arrange()
// Like I said, this does all the work. It's somewhat localized
// for each edge of each view as to reduce redundancy and allow
// for more flexibility in dependencies. It is still possible to
// have circular dependencies yet layout managers sitting on top
// of this manager should be intelligent to prevent this.
//
// This function generally just shifts each edge according to the
// relationship described in the WFrame class. Its adjustment is
// synchronized by the parentFrame parameter to ensure that
// everyone thinks 10% is X pixels...

void WEdgeManager::Arrange(WFrameEntry *entry,
								const ulong edge,
								const BRect& pFrame)
{
	BRect parentFrame = pFrame;
	if (entry == NULL) return;
	if (entry->view->Window() == NULL) return;
	
	BView *view = entry->view;	
	WFrame *frame = entry->frame;
	BRect oldFrame = view->Frame();
	BRect b = view->Bounds();
	//printf("Looking at the left side:\n");
	//printf("frame left = %d, bounds left = %d.\n", (int)oldFrame.left, (int)b.left);

	oldFrame.OffsetTo(0,0);
	
	if (view->Parent()) {
		parentFrame = (view->Parent()->Bounds());
		parentFrame.OffsetTo(0,0);
		//printf("Parent is %d,%d x %d,%d.\n",(int)parentFrame.left,(int)parentFrame.top,
		//		(int)parentFrame.right,(int)parentFrame.bottom);
	}
	else {
		parentFrame = view->Window()->Bounds(); 
		parentFrame.OffsetTo(0,0);
		//printf("Parent is %d,%d x %d,%d.\n",(int)parentFrame.left,(int)parentFrame.top,
		//		(int)parentFrame.right,(int)parentFrame.bottom);
	}
	
	WFrameEntry *sibEntry = 0;
	switch (edge) {
		case EDGE_LEFT:
		{
			//printf("Adjusting edge left. (relation = %d, style = %d)\n", frame->left_relation, frame->left_style);
			if (entry->isLeftSet) return;
			else entry->isLeftSet = true;

			switch(frame->left_relation)
			{
				case USE_NONE:
				{
					BRect r = entry->view->Frame();
					entry->newFrame.left = r.left;
					break;
				}
				case USE_DELTA:
					if (!entry->isRightSet) Arrange(entry,EDGE_RIGHT, parentFrame);
					entry->newFrame.left = entry->newFrame.right + frame->left_value;
					break;
				case USE_PARENT:
					//printf("\tUsing parent.\n");
					switch(frame->left_style)
					{
						case USE_LITERAL:
							entry->newFrame.left = frame->left_value ;
							break;
						case USE_PERCENT:
							entry->newFrame.left = (int)(ROUND_ERROR_CORRECT + (frame->left_value / 100 * parentFrame.right));
							break;
					}
					break;
				case USE_SIBLING:
					switch(frame->left_style)
					{
						sibEntry = FindEntry(frame->left_view);
						case EDGE_ADJACENT:
							if (sibEntry) {
								//printf("\t** Arrange: left according to sib at: %f\n",sibEntry->newFrame.right + 1 - frame->left_value);
								if (!sibEntry->isRightSet) Arrange(sibEntry, EDGE_RIGHT, parentFrame);
								entry->newFrame.left = sibEntry->newFrame.right + 1 + frame->left_value;
							}
							else {
								BRect r = frame->left_view->Frame();
								//r.OffsetTo(0,0);
								//printf("\t** Arrange: left NOT according to sib at: %f\n",r.right + 1 - frame->left_value);
								//printf("\t** r.right = %f.\n", r.right);
								//printf("\t** Frame Value = %f.\n", frame->left_value);
								entry->newFrame.left = r.right + 1 + frame->left_value;
							}
							break;
						case EDGE_OPPOSITE:

							if (sibEntry) {
								if (!sibEntry->isLeftSet) Arrange(sibEntry, EDGE_LEFT, parentFrame);
								entry->newFrame.left = sibEntry->newFrame.left + frame->left_value;
							}
							// changed Frame().right to Frame().left
							else {
								BRect r = frame->left_view->Frame();
								entry->newFrame.left = r.left + frame->left_value;
							}
							break;
						case USE_PERCENT:
							if (sibEntry) {
								if (!sibEntry->isLeftSet) Arrange(sibEntry, EDGE_LEFT, parentFrame);
								if (!sibEntry->isRightSet) Arrange(sibEntry, EDGE_RIGHT, parentFrame);
								entry->newFrame.left = (int)(ROUND_ERROR_CORRECT+sibEntry->newFrame.left +
															(frame->left_value / 100 * (sibEntry->newFrame.right - sibEntry->newFrame.left)));
							}
							else {
								BRect fL = frame->left_view->Bounds();
								fL.OffsetTo(0,0);
								//printf("Parent is %d,%d x %d,%d.\n",(int)parentFrame.left,(int)parentFrame.top,
								//	(int)parentFrame.right,(int)parentFrame.bottom);
								BRect r = frame->left_view->Frame();
								entry->newFrame.left = (int)(ROUND_ERROR_CORRECT+r.left+
															(frame->left_value / 100 * fL.right));
							}
							break;
					}
					break;
				default:
					//printf("What sort of layout behaviour I forget to code??\n");
					break;
			}
			break;
		}

		case EDGE_RIGHT:
		{
			//printf("Adjusting edge right. (relation = %d, style = %d)\n", frame->right_relation, frame->right_style);
			if (entry->isRightSet)  return;
			else entry->isRightSet = true;

			switch(frame->right_relation)
			{
				case USE_NONE:
					BRect r = entry->view->Frame();
					entry->newFrame.right = r.right;
					break;
				case USE_DELTA:
					if (!entry->isLeftSet) Arrange(entry,EDGE_LEFT, parentFrame);
					entry->newFrame.right = entry->newFrame.left + frame->right_value;
					break;
				case USE_PARENT:
					//printf("\tUsing parent.\n");
					switch(frame->right_style)
					{
						case USE_LITERAL:
						//printf("parentFrame.right = %f-%f.\n",parentFrame.right,frame->right_value); 
							entry->newFrame.right = /*parentFrame.right-*/frame->right_value;
							break;
						case USE_PERCENT:
							entry->newFrame.right = (int)(ROUND_ERROR_CORRECT+(frame->right_value / 100 * parentFrame.right));
							break;
					}
					break;
				case USE_SIBLING:
					//printf("Using sibling\n");
					//printf("Edge_Right = %s.\n",frame->right_view->Name());
					//printf("calling findentry\n");
					sibEntry = FindEntry(frame->right_view);
					switch(frame->right_style)
					{
						case EDGE_ADJACENT:
							//printf("Using edge adjacent\n");
							if (sibEntry) {
								if (!sibEntry->isLeftSet) Arrange(sibEntry, EDGE_LEFT, parentFrame);
								entry->newFrame.right = sibEntry->newFrame.left - 1 + frame->right_value;
							}
							else {
								BRect r = frame->right_view->Frame();
								r.OffsetTo(0,0);
								entry->newFrame.right = r.left - 1 + frame->right_value;
							}
							break;
						case EDGE_OPPOSITE:
							if (sibEntry) {
								if (!sibEntry->isRightSet) Arrange(sibEntry, EDGE_RIGHT, parentFrame);
								entry->newFrame.right = sibEntry->newFrame.right + frame->right_value;
							}
							else {
								BRect r = frame->right_view->Frame();
								r.OffsetTo(0,0);
								entry->newFrame.right = r.right + frame->right_value;
							}
							break;
						case USE_PERCENT:
							if (sibEntry) {
								if (!sibEntry->isLeftSet) Arrange(sibEntry, EDGE_LEFT, parentFrame);
								if (!sibEntry->isRightSet) Arrange(sibEntry, EDGE_RIGHT, parentFrame);
								entry->newFrame.right =  (int)(ROUND_ERROR_CORRECT+sibEntry->newFrame.left +
															(frame->right_value / 100 *(sibEntry->newFrame.right - sibEntry->newFrame.left)));
							}
							else {
								BRect fR = frame->right_view->Bounds();
								fR.OffsetTo(0,0);
								//printf("Parent is %d,%d x %d,%d.\n",(int)parentFrame.left,(int)parentFrame.top,
								//	(int)parentFrame.right,(int)parentFrame.bottom);
								BRect r = frame->right_view->Frame();
								r.OffsetTo(0,0);
								entry->newFrame.right =  (int)(ROUND_ERROR_CORRECT+r.left +
															(frame->right_value / 100 * (fR.right)));
							}
							break;
					}
					break;
				default:
					//printf("What sort of layout behaviour did I forget to code??\n");
					break;
			}
			break;
		}
		case EDGE_TOP:
		{
			//printf("Adjusting edge top. (relation = %d, style = %d)\n", frame->top_relation, frame->top_style);
			if (entry->isTopSet) return;
			else entry->isTopSet = true;

			switch(frame->top_relation)
			{
				case USE_NONE:
				{
					BRect r = entry->view->Frame();
					entry->newFrame.top = r.top;
					break;
				}
				case USE_DELTA:
					if (!entry->isBottomSet) Arrange(entry,EDGE_BOTTOM, parentFrame);
					entry->newFrame.top = entry->newFrame.bottom + frame->top_value;
					break;
				case USE_PARENT:
					//printf("\t** Using parent.\n");
					switch(frame->top_style)
					{
						case USE_LITERAL:
							entry->newFrame.top = frame->top_value;
							break;
						case USE_PERCENT:
							//printf("\t** Using percentage. value = %f, placing at %d.\n",frame->top_value, (int)(frame->top_value/100*parentFrame.bottom) );
							entry->newFrame.top = (int)(ROUND_ERROR_CORRECT+(frame->top_value / 100 * parentFrame.bottom));
							break;
					}
					break;
				case USE_SIBLING:
					switch(frame->top_style)
					{
						case EDGE_ADJACENT:
							sibEntry = FindEntry(frame->top_view);
							if (sibEntry) {
								if (!sibEntry->isBottomSet) Arrange(sibEntry, EDGE_BOTTOM, parentFrame);
								entry->newFrame.top = sibEntry->newFrame.bottom + 1 + frame->top_value;
							}
							else {
								BRect r = frame->top_view->Frame();
								r.OffsetTo(0,0);
								entry->newFrame.top = r.bottom + 1 + frame->top_value;
							}
							break;
						case EDGE_OPPOSITE:
							sibEntry = FindEntry(frame->top_view);
							if (sibEntry) {
								if (!sibEntry->isTopSet) Arrange(sibEntry, EDGE_TOP, parentFrame);
								entry->newFrame.top = sibEntry->newFrame.top + frame->top_value;
							}
							else {
								BRect r = frame->top_view->Frame();
								r.OffsetTo(0,0);
								entry->newFrame.top = r.bottom + frame->top_value;
							}
							break;
						case USE_PERCENT:
							sibEntry = FindEntry(frame->top_view);
							if (sibEntry) {
								if (!sibEntry->isTopSet) Arrange(sibEntry, EDGE_TOP, parentFrame);
								if (!sibEntry->isBottomSet) Arrange(sibEntry, EDGE_BOTTOM, parentFrame);
								entry->newFrame.top = (int)(ROUND_ERROR_CORRECT + sibEntry->newFrame.top +
														(frame->top_value / 100 * (sibEntry->newFrame.bottom - sibEntry->newFrame.top)));
							}
							else {
								BRect fT = frame->top_view->Bounds();
								fT.OffsetTo(0,0);
								//printf("Parent is %d,%d x %d,%d.\n",(int)parentFrame.left,(int)parentFrame.top,
								//	(int)parentFrame.right,(int)parentFrame.bottom);
								BRect r = frame->top_view->Frame();
								r.OffsetTo(0,0);
								entry->newFrame.top = (int)(ROUND_ERROR_CORRECT + r.top +
															(frame->top_value / 100 * fT.bottom));
							}
							break;
					}
					break;
				default:
					//printf("What sort of layout behaviour I forget to code??\n");
					break;
			}
			break;
		}

		case EDGE_BOTTOM:
		{
			//printf("Adjusting edge bottom. (relation = %d, style = %d)\n", frame->bottom_relation, frame->bottom_style);
			if (entry->isBottomSet)  return;
			else entry->isBottomSet = true;

			switch(frame->bottom_relation)
			{
				case USE_NONE:
				{
					BRect r = entry->view->Frame();
					entry->newFrame.bottom = r.bottom;
					break;
				}
				case USE_DELTA:
					if (!entry->isTopSet) Arrange(entry,EDGE_TOP, parentFrame);
					entry->newFrame.bottom = entry->newFrame.top + frame->bottom_value;
					break;
				case USE_PARENT:
					//printf("\tUsing parent.\n");
					switch(frame->bottom_style)
					{
						case USE_LITERAL:
							//printf("Bottom = %d - %d. Why?", (int)parentFrame.bottom, (int)frame->bottom_value);
							entry->newFrame.bottom = /*parentFrame.bottom-*/frame->bottom_value;
							break;
						case USE_PERCENT:
							entry->newFrame.bottom = (int)(ROUND_ERROR_CORRECT+(frame->bottom_value / 100 * parentFrame.bottom));
							break;
					}
					break;
				case USE_SIBLING:
					switch(frame->bottom_style)
					{
						case EDGE_ADJACENT:
							sibEntry = FindEntry(frame->bottom_view);
							if (sibEntry) {
								if (!sibEntry->isTopSet) Arrange(sibEntry, EDGE_TOP, parentFrame);
								entry->newFrame.bottom = sibEntry->newFrame.top - 1 + frame->bottom_value;
							}
							else {
								BRect r = frame->bottom_view->Frame();
								r.OffsetTo(0,0);
								entry->newFrame.bottom = r.top - 1 + frame->bottom_value;
							}
							break;
						case EDGE_OPPOSITE:
							sibEntry = FindEntry(frame->bottom_view);
							if (sibEntry) {
								if (!sibEntry->isBottomSet) Arrange(sibEntry, EDGE_BOTTOM, parentFrame);
								entry->newFrame.bottom = sibEntry->newFrame.bottom - frame->bottom_value;
							}
							else {
								BRect r = frame->bottom_view->Frame();
								r.OffsetTo(0,0);
								entry->newFrame.bottom = r.bottom - frame->bottom_value;
							}
							break;
						case USE_PERCENT:
							sibEntry = FindEntry(frame->bottom_view);
							if (sibEntry) {
								if (!sibEntry->isTopSet) Arrange(sibEntry, EDGE_TOP, parentFrame);
								if (!sibEntry->isBottomSet) Arrange(sibEntry, EDGE_BOTTOM, parentFrame);
								entry->newFrame.bottom =  (int)(ROUND_ERROR_CORRECT+sibEntry->newFrame.top +
															(frame->bottom_value / 100 *(sibEntry->newFrame.bottom - sibEntry->newFrame.top)));
							}
							else {
								BRect fB = frame->bottom_view->Bounds();
								fB.OffsetTo(0,0);
								//printf("Parent is %d,%d x %d,%d.\n",(int)parentFrame.left,(int)parentFrame.top,
								//	(int)parentFrame.right,(int)parentFrame.bottom);
								BRect r = frame->bottom_view->Frame();
								r.OffsetTo(0,0);
								entry->newFrame.bottom =  (int)(ROUND_ERROR_CORRECT+r.top +
																(frame->bottom_value / 100 * (fB.bottom)));
							}
							break;
					}
					break;
				default:
					//printf("What sort of layout behaviour did I forget to code??\n");
					break;
			}
			break;
		}
	}
}